<?php declare(strict_types=1);
/*
 * This file is part of PHPUnit.
 *
 * (c) Sebastian Bergmann <sebastian@phpunit.de>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
namespace PHPUnit\TestFixture;

use PHPUnit\Framework\Attributes\CoversNothing;
use PHPUnit\Framework\Attributes\DataProvider;
use PHPUnit\Framework\Attributes\Small;
use PHPUnit\Framework\TestCase;

#[Small]
#[CoversNothing]
final class SmallTestWithDataProvider extends TestCase
{
    public static function provideData(): array
    {
        return [
            [true],
            [false],
        ];
    }

    #[DataProvider('provideData')]
    public function testWithProvider(bool $value): void
    {
        $this->assertTrue(true);
    }

    public function testWithoutProvider(): void
    {
        $this->assertTrue(true);
    }
}
