<?php declare(strict_types=1);
namespace PHPUnit\TestFixture\ErrorHandler\DeprecationTriggerFunction;

use const E_USER_DEPRECATED;
use function trigger_error;

function trigger_deprecation(string $message): void
{
    @trigger_error($message, E_USER_DEPRECATED);
}
