<?php declare(strict_types=1);
/*
 * This file is part of PHPUnit.
 *
 * (c) Sebastian Bergmann <sebastian@phpunit.de>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
namespace PHPUnit\TestFixture\Issue4923;

use PHPUnit\Framework\TestCase;
use stdClass;

final class Issue4923Test extends TestCase
{
    public function testOne(): void
    {
        $obj = new stdClass;

        $a = [123, $obj, 'hello'];
        $b = [$obj, 'hello', 123];

        $this->assertEqualsCanonicalizing($a, $b);
    }
}
